/****************************************************************************
*                   kepler.c
*
*  This module implements routines for quasi constructive solid geometry.
*
*  from Persistence of Vision(tm) Ray Tracer
*  Copyright 2000 J. Grimbert
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file.
*  If POVLEGAL.DOC is not available or for more info please contact the POV-Ray
*  Team Coordinator by email to team-coord@povray.org or visit us on the web at
*  http://www.povray.org. The latest version of POV-Ray may be found at this site.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
*****************************************************************************/

#include "frame.h"
#include "povray.h"
#include "vector.h"
#include "povproto.h"
#include "bbox.h"
#include "csg.h"
#include "hfield.h"
#include "matrices.h"
#include "objects.h"
#include "planes.h"
#include "quadrics.h"
#include "kepler.h"



/*****************************************************************************
* Local preprocessor defines
******************************************************************************/




/*****************************************************************************
* Static functions
******************************************************************************/

static int All_KEPLER_Intersections (OBJECT *Object, RAY *Ray, ISTACK *Depth_Stack);
static int Inside_KEPLER (VECTOR point, OBJECT *Object);
static int Inside_KEPLER2 (VECTOR point, OBJECT *Object);

static KEPLER *Copy_KEPLER (OBJECT *Object);
static void Translate_KEPLER (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Rotate_KEPLER (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Scale_KEPLER (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans);
static void Transform_KEPLER (OBJECT *Object, TRANSFORM *Trans);
static void Destroy_KEPLER (OBJECT *Object);
static void Invert_KEPLER (OBJECT *Object);


/*****************************************************************************
* Local variables
******************************************************************************/

METHODS KEPLER_Methods =
{
  All_KEPLER_Intersections,
  Inside_KEPLER, NULL /*Normal*/,
  (COPY_METHOD)Copy_KEPLER,
  Translate_KEPLER, Rotate_KEPLER,
  Scale_KEPLER, Transform_KEPLER, Invert_KEPLER, Destroy_KEPLER
};

METHODS KEPLER2_Methods =
{
  All_KEPLER_Intersections,
  Inside_KEPLER2, NULL /*Normal*/,
  (COPY_METHOD)Copy_KEPLER,
  Translate_KEPLER, Rotate_KEPLER,
  Scale_KEPLER, Transform_KEPLER, Invert_KEPLER, Destroy_KEPLER
};

/*****************************************************************************
*
* FUNCTION
*
*   All_KEPLER_Intersections
*
* INPUT
*   
* OUTPUT
*   
* RETURNS
*   
* AUTHOR
*
*   J. Grimbert
*   
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static int All_KEPLER_Intersections (OBJECT *Object, RAY *Ray, ISTACK *Depth_Stack)
{
  int Found, Count;
  OBJECT *Current_Sib, *Inside_Sib;
  ISTACK *Local_Stack;
  INTERSECTION *Sibling_Intersection;

  Increase_Counter(stats[Ray_KEPLER_Tests]);

  Found = FALSE;

  Local_Stack = open_istack();
  for ( Current_Sib = ((KEPLER *)Object)->Children; 
        Current_Sib != NULL; Current_Sib = Current_Sib->Sibling)
  {
    if (Ray_In_Bound (Ray, Current_Sib->Bound))
    {
      if (All_Intersections (Current_Sib, Ray, Local_Stack))
      {
        while ((Sibling_Intersection = pop_entry(Local_Stack)) != NULL)
        {
          Count = 1; /* at least this child */
          for (Inside_Sib = ((KEPLER *)Object)->Children; Inside_Sib != NULL; Inside_Sib = 
 Inside_Sib->Sibling)
          {
             if (Inside_Sib != Current_Sib)
             {
                if (Inside_Object(Sibling_Intersection->IPoint, Inside_Sib))
                {
                   Count++;
                }
             }
          }
          if ((Count >= ((KEPLER *)Object)->Minimal)
            &&((  !(((KEPLER *)Object)->Maximal ) )
              || ( Count <= ((KEPLER *)Object)->Maximal)
              )
             )
          {
             if (Point_In_Clip (Sibling_Intersection->IPoint, Object->Clip))
             {
                 push_copy (Depth_Stack, Sibling_Intersection);
                 Found = TRUE;
             }
          }
        }
      }
    }
  }

    close_istack (Local_Stack);

  if (Found)
  {
    Increase_Counter(stats[Ray_KEPLER_Tests_Succeeded]);
  }

  return (Found);
}

/*****************************************************************************
*
* FUNCTION
*
*   Inside_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static int Inside_KEPLER (VECTOR IPoint, OBJECT *Object)
{
  OBJECT *Current_Sib;
  int Count = 0;

  for (Current_Sib = ((KEPLER *)Object)->Children; Current_Sib != NULL; Current_Sib = Current_Sib->Sibling)
  {
    if (Inside_Object (IPoint, Current_Sib))
    {
      Count++;
      if (Count >= ((KEPLER *)Object)->Minimal  )
         return (TRUE);
    }
  }
  return (FALSE);
}


/*****************************************************************************
*
* FUNCTION
*
*   Inside_KEPLER2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static int Inside_KEPLER2 (VECTOR IPoint, OBJECT *Object)
{
  OBJECT *Current_Sib;
  int Count = 0;

  for (Current_Sib = ((KEPLER *)Object)->Children; Current_Sib != NULL; Current_Sib = Current_Sib->Sibling)
  {
    if (Inside_Object (IPoint, Current_Sib))
    {
      Count++;
    }
  }

  if ((Count < ((KEPLER *)Object)->Maximal )
     &&(Count >= ((KEPLER *)Object)->Minimal )
     )
      return (TRUE);
  return (FALSE);
}



/*****************************************************************************
*
* FUNCTION
*
*   Translate_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Translate_KEPLER (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((KEPLER *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Translate_Object(Sib, Vector, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Rotate_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Rotate_KEPLER (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((KEPLER *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Rotate_Object(Sib, Vector, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Scale_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Scale_KEPLER (OBJECT *Object, VECTOR Vector, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((KEPLER *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Scale_Object(Sib, Vector, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Transform_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Transform_KEPLER (OBJECT *Object, TRANSFORM *Trans)
{
  OBJECT *Sib;

  for (Sib = ((KEPLER *) Object)->Children; Sib != NULL; Sib = Sib->Sibling)
  {
    Transform_Object (Sib, Trans);
  }

  Recompute_BBox(&Object->BBox, Trans);
}



/*****************************************************************************
*
* FUNCTION
*
*   Invert_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Invert_KEPLER (OBJECT *Object)
{
  Invert_Flag(Object, INVERTED_FLAG);
}



/*****************************************************************************
*
* FUNCTION
*
*   Create_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

KEPLER *Create_KEPLER ()
{
  KEPLER *New;

  New = (KEPLER *)POV_MALLOC(sizeof (KEPLER), "kepler");

  INIT_OBJECT_FIELDS(New, KEPLER_OBJECT, &KEPLER_Methods)

  New->Children = NULL;

  return (New);
  
}


/*****************************************************************************
*
* FUNCTION
*
*   Create_KEPLER2
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

KEPLER *Create_KEPLER2 ()
{
  KEPLER *New;

  New = (KEPLER *)POV_MALLOC(sizeof (KEPLER), "kepler2");

  INIT_OBJECT_FIELDS(New, KEPLER_OBJECT, &KEPLER2_Methods)

  New->Children = NULL;

  return (New);
  
}




/*****************************************************************************
*
* FUNCTION
*
*   Copy_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static KEPLER *Copy_KEPLER (OBJECT *Object)
{
  KEPLER *New;
  OBJECT *Old_Sib, *New_Sib, *Prev_Sib;

  New = (KEPLER *)POV_MALLOC(sizeof (KEPLER), "kepler");

  *New = *(KEPLER *)Object;

  New->Children = Prev_Sib = NULL;

  for (Old_Sib = ((KEPLER *)Object)->Children; Old_Sib != NULL; Old_Sib = Old_Sib->Sibling)
  {
    New_Sib = Copy_Object (Old_Sib);

    if (New->Children == NULL)
    {
      New->Children = New_Sib;
    }

    if (Prev_Sib != NULL)
    {
      Prev_Sib->Sibling = New_Sib;
    }

    Prev_Sib = New_Sib;
  }

  return (New);
}



/*****************************************************************************
*
* FUNCTION
*
*   Destroy_KEPLER
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

static void Destroy_KEPLER (OBJECT *Object)
{
  Destroy_Object (((KEPLER *) Object)->Children);

  POV_FREE (Object);
}



/*****************************************************************************
*
* FUNCTION
*
*   Compute_KEPLER_BBox
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   J. Grimbert
*
* DESCRIPTION
*
*   as usual
*
* CHANGES
*
*   -
*
******************************************************************************/

void Compute_KEPLER_BBox (OBJECT *Object)
{
  DBL Old_Volume, New_Volume;
  VECTOR NewMin, NewMax, TmpMin, TmpMax, Min, Max;
  OBJECT *Sib;
  int Count;

  /* 
  ** Worst bounding box : union of all the bounding box
  ** needed for the Minimal == 1
  **
  ** In fact, the Minimal should be used here to choose between
  **-the union of all bounding boxes of all objects (1)
  **-the union of all bounding boxes of all the intersections of 
  **  every couples of objects (2)
  ** and so on until Minimal is equal to the number of objects,
  ** where the bounding box is the intersection of all bounding boxes.
  **
  ** but that's far too complex for now
  ** 
  ** at least we can skip the (Minimal-1) first bounding box(es)
  */

  Make_Vector(NewMin,  BOUND_HUGE,  BOUND_HUGE,  BOUND_HUGE);
  Make_Vector(NewMax, -BOUND_HUGE, -BOUND_HUGE, -BOUND_HUGE);

  for (Sib = ((KEPLER *)Object)->Children, Count = ((KEPLER *)Object)->Minimal; 
   ( Sib != NULL)&&(Count > 1); Count--, Sib = Sib->Sibling);
  for (; Sib != NULL; Sib = Sib->Sibling)
  {
    Make_min_max_from_BBox(TmpMin, TmpMax, Sib->BBox);

    NewMin[X] = min(NewMin[X], TmpMin[X]);
    NewMin[Y] = min(NewMin[Y], TmpMin[Y]);
    NewMin[Z] = min(NewMin[Z], TmpMin[Z]);
    NewMax[X] = max(NewMax[X], TmpMax[X]);
    NewMax[Y] = max(NewMax[Y], TmpMax[Y]);
    NewMax[Z] = max(NewMax[Z], TmpMax[Z]);
  }

  if ((NewMin[X] > NewMax[X]) || (NewMin[Y] > NewMax[Y]) || (NewMin[Z] > NewMax[Z]))
  {
    Warning(0.0, "Degenerate KEPLER bounding box (not used!).\n");
  }
  else
  {
    New_Volume = (NewMax[X] - NewMin[X]) * (NewMax[Y] - NewMin[Y]) * (NewMax[Z] - NewMin[Z]);

    BOUNDS_VOLUME(Old_Volume, Object->BBox);

    if (New_Volume < Old_Volume)
    {
      Make_BBox_from_min_max(Object->BBox, NewMin, NewMax);

      /* Beware of bounding boxes to large. */

      if ((Object->BBox.Lengths[X] > CRITICAL_LENGTH) ||
          (Object->BBox.Lengths[Y] > CRITICAL_LENGTH) ||
          (Object->BBox.Lengths[Z] > CRITICAL_LENGTH))
      {
        Make_BBox(Object->BBox, -BOUND_HUGE/2, -BOUND_HUGE/2, -BOUND_HUGE/2,
          BOUND_HUGE, BOUND_HUGE, BOUND_HUGE);
      }
    }
  }
}
